// No copyright
#pragma once

#include <assert.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>


#define ACG_CHUNK_BLOB 0x00

#define ACG_CHUNK_BEGIN 0x01
#define ACG_CHUNK_END   0x02
#define ACG_CHUNK_CHECK 0x03


static inline uint64_t acg_util_fullbits(uint8_t bits) {
  return (((uint64_t) 1U) << bits) - 1U;
}
static inline uint64_t acg_util_umin(uint64_t a, uint64_t b) {
  return a > b? b: a;
}


struct acg_stream {
  uint8_t* buffer;
  uint64_t size_bytes;

  uint64_t cursor_bits;
};

static inline bool acg_stream_write_bool(struct acg_stream* s, bool v) {
  if (s->size_bytes <= s->cursor_bits) {
    return false;
  }
  uint64_t* const c = &s->cursor_bits;
  s->buffer[*c/8U] |= (!!v) << (*c%8U);
  ++*c;
  return true;
}
static inline bool acg_stream_write_uint(struct acg_stream* s, uint64_t v, uint8_t unit_bits) {
  assert(unit_bits >= 2U);
  assert(unit_bits <= 65U);

  uint8_t*  const buf = s->buffer;
  uint64_t* const c   = &s->cursor_bits;

  const uint8_t unit_data_bits = unit_bits - 1U;
  while (v) {
    const uint64_t next_v = v >> unit_data_bits;
    acg_stream_write_bool(s, !!next_v);

    const uint8_t frag_bits = acg_util_umin(unit_data_bits, (8U - *c%8U)%8U);
    if (frag_bits > 0U) {
      buf[*c/8U] |= (v & acg_util_fullbits(frag_bits)) << (*c%8U);
      *c += frag_bits;
    }

    const uint8_t  remain_bits = unit_data_bits - frag_bits;
    const uint64_t v_part      = (v >> frag_bits) & acg_util_fullbits(remain_bits);
    memcpy(&buf[*c/8U], &v_part, (remain_bits + 7U) / 8U);
    *c += remain_bits;

    v = next_v;
  }
  return true;
}
static inline bool acg_stream_write_int(struct acg_stream* s, int64_t v, uint8_t unit_bits) {
  assert(unit_bits > 0);
  const uint64_t uv = v >= 0? ((uint64_t) v): (~((uint64_t) -v) + 1);
  return acg_stream_write_uint(s, uv, unit_bits);
}
